#pylint: disable=E0401
"""
Test SocketIO abstract source.
"""
import asyncio
import json
from unittest import mock
from unittest import TestCase

import websockets

from aiopype import SyncProtocol
from aiopype.sources import socketio


class TestSocketIOPacket(TestCase):
  """
  Test huobi packet wrapper.
  """
  def test_packet_str(self):
    packet = socketio.SocketIOPacket('event', 'test', {'some': 'data'})

    packet_str = str(packet)
    packet_message = json.loads(packet_str.split('5:::')[1])

    self.assertTrue(packet_str.startswith('5:::'))
    self.assertEqual(packet_message, {"name": "test", "args": [{"some": "data"}]})


class TestSocketIOSource(TestCase):
  async def sleeper(self, time):
    await asyncio.sleep(time)

  def test_on_error(self):
    mock_handler = SyncProtocol()
    source = socketio.SocketIOSource('test', mock_handler)
    source.websocket = mock.MagicMock()

    loop = asyncio.get_event_loop()
    exception = None
    try:
      loop.run_until_complete(source.on_error('abc'))
    except Exception as error:
      exception = error

    self.assertTrue(source.websocket.close.called)
    self.assertEqual(type(exception), websockets.ConnectionClosed)

  def test_on_disconnect(self):
    mock_handler = SyncProtocol()
    source = socketio.SocketIOSource('test', mock_handler)
    source.max_failures = 0
    loop = asyncio.get_event_loop()
    loop.run_until_complete(source.on_disconnect('test'))
    self.assertFalse(source.running)

  @mock.patch('aiopype.sources.socketio.aiohttp')
  def test_get_url(self, aiohttp_mock):
    mock_handler = SyncProtocol()
    source = socketio.SocketIOSource('test', mock_handler)

    json_future = asyncio.Future()
    json_future.set_result("123123:1231231")
    http_response = mock.Mock()
    http_response.text = mock.Mock(return_value = json_future)

    response_future = asyncio.Future()
    response_future.set_result(http_response)
    http_client = mock.Mock()
    http_client.post = mock.Mock(return_value = response_future)

    session = mock.MagicMock()
    session.__enter__ = mock.Mock(return_value = http_client)
    aiohttp_mock.ClientSession = mock.Mock(return_value = session)

    loop = asyncio.get_event_loop()
    url = loop.run_until_complete(source.get_url())

    self.assertEqual(url, 'ws://localhost/socket.io/1/websocket/123123')

  def test_start(self):
    mock_handler = SyncProtocol()
    source = socketio.SocketIOSource('test', mock_handler)
    source.connect = mock.Mock(return_value = 'test')
    watch_future = asyncio.Future()
    watch_future.set_result(None)
    source.watchdog = mock.Mock(return_value = watch_future)
    socketio.SocketIOSource.running = mock.PropertyMock(side_effect = [True, True, False])
    exception = None

    try:
      loop = asyncio.get_event_loop()
      loop.run_until_complete(source.start())
    except Exception as error:
      exception = error

    self.assertEqual(str(exception), 'Disconnected too many times, stopped')
    self.assertTrue(source.done)
    source.watchdog.assert_called_with('test')
    socketio.SocketIOSource.running = True

  def test_watchdog(self):
    mock_handler = SyncProtocol()
    source = socketio.SocketIOSource('test', mock_handler)
    done_future = asyncio.Future()
    done_future.set_result(None)
    source.emit_async = mock.Mock(return_value = done_future)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(source.watchdog(self.sleeper(1)))

    source.emit_async.assert_called_with('disconnect', 'Lost heartbeat')

    source.heartbeat_timeout = 2
    loop = asyncio.get_event_loop()
    coro = self.sleeper(10)
    loop.run_until_complete(source.watchdog(coro))

    source.emit_async.assert_called_with('disconnect', 'Lost heartbeat')

  def test_handle_heartbeat(self):
    mock_handler = SyncProtocol()
    source = socketio.SocketIOSource('test', mock_handler)
    source.websocket = mock.Mock()
    done_future = asyncio.Future()
    done_future.set_result(None)
    source.websocket.send = mock.Mock(return_value = done_future)
    source.last_heartbeat = None

    loop = asyncio.get_event_loop()
    loop.run_until_complete(source.handle_heartbeat('test'))

    source.websocket.send.assert_called_with('2::')
    self.assertTrue(source.last_heartbeat)



#pylint: disable=E0401, R0201
"""
Test flow controller
"""
from unittest import mock
from unittest import TestCase

from aiopype.protocol import SyncProtocol
from aiopype.flow import Flow
from aiopype.manager import Manager

def zero_error():
  return 1 / 0

def run_loop(loop):
  handle = loop.call_soon(zero_error)
  loop._run_once()
  return handle


class TestInit(TestCase):
  def setUp(self):
    class TestManager(Manager):
      name = 'test'
      async def stop(self):
        pass

    self.test_manager = TestManager
    config_mock = mock.Mock()
    config_mock.FLOWS = ['test']
    config_mock.RAVEN_DSN = None
    self.flow = Flow(config_mock)
    self.flow.running = True

  def test_start(self):
    loop = mock.Mock()
    loop.run_forever = mock.Mock()
    self.test_manager.start = mock.Mock(return_value = None)
    self.flow.finished = mock.Mock(return_value = "check termination condition perodically")

    self.flow.loop = loop
    self.flow.start()
    self.test_manager.start.assert_called_with(loop)
    loop.run_until_complete.assert_called_with("check termination condition perodically")

  def test_check_managers(self):
    """
    Test if recovers from failed manager.
    """
    self.test_manager.done = mock.Mock(side_effect = [True, False])
    self.test_manager.should_restart = mock.Mock(return_value = True)
    self.test_manager.start = mock.Mock(return_value = None)
    self.flow.managers = [self.test_manager()]
    self.flow.handler = SyncProtocol()

    self.flow.loop.run_until_complete(self.flow.check_managers())
    self.assertTrue(1, len(self.flow.managers))
    self.test_manager.start.assert_called_with(self.flow.loop)

  def test_no_restart(self):
    """
    Test if manager with run_always set to false is ignored.
    """
    manager = self.test_manager()
    manager.should_restart = mock.Mock(return_value = False)
    manager.done = mock.Mock(return_value = True)

    self.flow.managers = [manager]

    self.flow.loop.run_until_complete(self.flow.check_managers())
    self.assertFalse(self.flow.managers)

  def test_handle_exceptions(self):
    config_mock = mock.Mock()
    config_mock.FLOWS = ['test']
    config_mock.RAVEN_DSN = None
    self.flow = Flow(config_mock)
    self.flow.running = True

    self.flow.logger.info = mock.Mock()
    run_loop(self.flow.loop)

    args = self.flow.logger.info.call_args

    self.assertTrue(args)

    self.assertEqual(args[1]['extra']['info'][0], ZeroDivisionError)
    self.assertTrue(isinstance(args[1]['extra']['info'][1], ZeroDivisionError))

  def test_raven_capture(self):
    config_mock = mock.Mock()
    config_mock.FLOWS = ['test']
    config_mock.RAVEN_DSN = None
    self.flow = Flow(config_mock)
    self.flow.running = True

    self.flow.raven = mock.Mock()
    self.flow.raven.captureException = mock.Mock()

    run_loop(self.flow.loop)

    self.assertTrue(self.flow.raven.captureException.called)



"""
Processor module.

Responsible for processing incoming data and produce events.
"""

import logging

from aiopype.protocol import EventMixin


class ProcessorRegistry(type):
  """
  Processors metaclass with registry.
  """

  REGISTRY = {}

  def __new__(mcs, name, bases, attrs):
    """ @param name: Name of the class
        @param bases: Base classes (tuple)
        @param attrs: Attributes defined for the class
    """
    new_cls = type.__new__(mcs, name, bases, attrs)
    mcs.REGISTRY[new_cls.__module__ + '.' + new_cls.__name__] = new_cls

    return new_cls

class Processor(EventMixin, metaclass = ProcessorRegistry):
  """
  Processor object.
  """
  def __init__(self, name, handler):
    """
    Constructor.
      @param name: the unique identifier for this processor inside a manager
      @param handler: the central event handler
    """
    self.done = False
    self.name = name
    self.handler = handler
    self.logger = logging.getLogger(name)

  def start(self, loop):
    """
    Method to kickstart sources.

    Must be overriden by source processors.
    """
    raise NotImplementedError()

  def is_done(self):
    """
    Determines if a processor has for some reason stopped.
    """
    return self.done



from .pusher import PusherClientSource
from .rest import RestSource
from .schedulers import CronJob, OffsetCronJob, HistoryTimeEmitter
from .socketio import SocketIOPacket, SocketIOPacketException, SocketIOSource
from .websocket import WebsocketSource



#pylint: disable=E0401
"""
Test processor.
"""
import asyncio
from unittest import TestCase

from aiopype import SyncProtocol
from aiopype.processor import Processor


class TestProcessor(TestCase):
  def test_emit_async(self):
    """
    Test processor's emit_async.
    """
    self.async_called = False
    self.sync_called = True

    async def async_handler():
      """
      Async handler.
      """
      await asyncio.sleep(1)
      self.async_called = True

    def sync_handler():
      """
      Sync handler.
      """
      self.sync_called = True

    async def run_emit():
      handled = await event_emitter.emit_async('test')
      self.assertTrue(handled)

    event_emitter = Processor('test', SyncProtocol())

    event_emitter.once('test', async_handler)
    event_emitter.once('test', sync_handler)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_emit())

    self.assertTrue(self.async_called)
    self.assertTrue(self.sync_called)

  def test_is_done(self):
    processor = Processor('test', SyncProtocol())
    self.assertFalse(processor.is_done())

    processor.done = True
    self.assertTrue(processor.is_done())

  def test_exception_without_handler(self):
    """
    Test exception without handler.
    """
    event_emitter = Processor('test', SyncProtocol())
    exception = None

    async def run_emit():
      await event_emitter.emit_async('error')

    try:
      loop = asyncio.get_event_loop()

      loop.run_until_complete(run_emit())
    except Exception as err:
      exception = err

    self.assertEqual(str(exception), 'Uncaught error event.')



"""
Messaging mixin for processors.
"""
class EventMixin(object):
  """
  Messaging Mixin.
  """
  def on(self, event, func):
    """
    Add handler for event.
    """
    self.handler.on(self.name + '__' + event, func)

  def once(self, event, func):
    """
    Add handler for one time only event.
    """
    self.handler.once(self.name + '__' + event, func)

  def emit(self, event, *args, **kwargs):
    """
    Trigger an event.
    """
    return self.handler.emit(self.name + '__' + event, *args, **kwargs)

  async def emit_async(self, event, *args, **kwargs):
    """
    Trigger an event asynchrounously.

    Returns a coroutine.
    """
    return await self.handler.emit_async(self.name + '__' + event, *args, **kwargs)

  def remove_all_listeners(self, event = None):
    """
    Remove all listeners attached to self.
    """
    if event:
      events = [self.name + '__' + event]
    else:
      events = [event for event in self.handler.get_events() if event.startswith(self.name + '__')]

    for event in events:
      self.handler.remove_all_listeners(event)

  def listeners(self, event):
    """
    Returns the list of all listeners registered to the event.
    """
    return self.handler.listeners(self.name + '__' + event)



import asyncio
import logging
from datetime import datetime, timedelta

import pusherclient

from aiopype.processor import Processor


class PusherClient(pusherclient.Pusher):
  """
  pusherclient.Pusher wrapper with error handling.

  Wraps calls to event handlers around try catch in order to capture exceptions
  to sentry. This is a workaround due to the threaded nature of PusherClient.
  """
  def __init__(self, *args, **kwargs):
    self.exception = None
    self.logger = logging.getLogger('pusherclient')
    kwargs['log_level'] = self.logger.getEffectiveLevel()

    super(PusherClient, self).__init__(*args, **kwargs)

  def _connection_handler(self, *args, **kwargs):
    try:
      super(PusherClient, self)._connection_handler(*args, **kwargs)
    except Exception as err:
      self.exception = err


class PusherClientSource(Processor):
  """
  Pusher Client source.

  Uses thread based pusherclient module, and monitors if the client is running as expected.
  """
  def __init__(self, name, handler, **kwargs):
    super().__init__(name, handler)
    self.done = False
    self.heartbeat_timeout = kwargs.get('heartbeat_timeout', 30)
    self.last_heartbeat = datetime.now()
    self.logger = logging.getLogger(name)
    self.pusher = PusherClient(kwargs.get('pusher', ''))

  async def start(self):
    """
    Initializes the Pusher client and exception / disconnected / heartbeat_timeout handlers.
    """
    self.pusher.connection.bind('pusher:connection_established', self.connect_handler)
    self.pusher.connect()

    try:
      while True:
        # Check pusher.exception and raise it.
        if self.pusher.exception:
          self.done = True
          raise self.pusher.exception

        # Check pusher.connection state.
        if self.pusher.connection.state in ('failed', 'disconnected'):
          self.done = True
          raise Exception("Connection to pusherclient lost {}".format(self.pusher.connection.state))

        if datetime.now() - self.last_heartbeat > timedelta(seconds = self.heartbeat_timeout):
          self.done = True
          raise Exception('Lost heartbeat for more than threshold value')

        await asyncio.sleep(0.5)
    finally:
      self.done = True

  def connect_handler(self, _):
    """
    The connection_established handler.

    Should subscribe to channels here.
    """
    pass



#pylint: disable=E0401, R0201
import asyncio
from unittest import mock
from unittest import TestCase

from aiopype.sources import RestSource


class TestRestSource(TestCase):
  def setUp(self):
    options = {
      'request_interval': 0.1
    }
    self.mock_handler = mock.MagicMock()
    self.source = RestSource('test', self.mock_handler,'test', options)
    self.loop = asyncio.get_event_loop()

  @mock.patch('aiopype.sources.rest.RestSource.running', new_callable = mock.PropertyMock)
  @mock.patch('aiopype.sources.rest.RestSource.get_data')
  def test_get_data_invoked(self, get_data_mock, running_mock):

    running_mock.side_effect = [True, True, False]
    future = asyncio.Future(loop = self.loop)
    future.set_result(None)
    get_data_mock.return_value = future

    async def test():
      await self.source.start()
      self.assertTrue(get_data_mock.called)

    self.loop.run_until_complete(test())

  @mock.patch('aiopype.sources.rest.RestSource.emit')
  @mock.patch('aiopype.sources.rest.aiohttp.request')
  def test_data_emit(self, request_mock, emit_mock):

    async def test():
      future_json = asyncio.Future(loop = self.loop)
      future_json.set_result({'some': 'data'})
      response = mock.Mock()
      response.headers = {'date': 'Thu, 25 Feb 2016 17:46:46 UTC'}
      response.json = mock.Mock(return_value = future_json)

      response_future = asyncio.Future(loop = self.loop)
      response_future.set_result(response)
      request_mock.return_value = response_future

      await self.source.get_data()

    self.loop.run_until_complete(test())
    emit_mock.assert_called_with('data', {'some': 'data'}, 1456422406)



import asyncio
import json
import logging

import websockets

from aiopype.processor import Processor


class WebsocketSource(Processor):
  url = None

  def __init__(self, name, handler, *args, **kwargs):
    super().__init__(name, handler)
    self.done = None
    self.loop = asyncio.get_event_loop()
    self.logger = logging.getLogger(name)
    self.running = True
    self.websocket = None
    self.failures = 0
    self.max_failures = kwargs.get('exception_threshold', 10)
    self.heartbeat_timeout = kwargs.get('heartbeat_timeout', 30)
    self.on('connected', self.on_connect)
    self.on('disconnected', self.on_lost_connection)

  async def on_connect(self):
    pass

  async def on_lost_connection(self, reason = 'Websocket Closed'):
    self.logger.warning('Disconnected', extra = {'reason': reason})
    self.failures += 1

    if self.failures > self.max_failures:
      self.running = False

  async def start(self):
    try:
      while self.running:
        await self.watchdog(self.connect())
    finally:
      self.done = True
      raise Exception('Disconnected too many times, stopped')

  async def connect(self):
    try:
      self.logger.debug('Connecting')
      async with websockets.connect(self.url) as self.websocket:
        self.logger.debug('Connected to websocket {}'.format(self.url))
        await self.emit_async('connected')

        while self.running:
          data_raw = await self.websocket.recv()
          received = self.parse(data_raw)
          await self.emit_async('message', received)
          self.failures = 0

    except (websockets.ConnectionClosed, ConnectionResetError, websockets.InvalidHandshake):
      await self.emit_async('disconnected')

  async def watchdog(self, coro):
    main_loop = asyncio.ensure_future(coro, loop = self.loop)
    websocket_timeout = 0

    while not self.websocket:
      if main_loop.done():
        await self.emit_async('disconnected', 'Service failed')

        return

      await asyncio.sleep(1)
      websocket_timeout += 1

      if websocket_timeout > self.heartbeat_timeout:
        main_loop.cancel()
        await self.emit_async('disconnected', 'Unable to connect')

        return

    while not main_loop.done():
      pong = await self.websocket.ping()
      done, pending = await asyncio.wait([main_loop, pong], timeout = self.heartbeat_timeout)

      if pong not in done:
        main_loop.cancel()
        await self.emit_async('disconnected', 'Lost heartbeat')

        return

      if main_loop in done:
        self.logger.debug(str(main_loop))
        await self.emit_async('disconnected', 'Service failed')

        return

  @staticmethod
  def parse(message):
    return json.loads(message)



"""
Cron Scheduler.
"""
import asyncio
from datetime import datetime

from aiopype.processor import Processor
from .time import get_rounded_timestamp

class CronJob(Processor):
  def __init__(self, name, handler, frequency, interval):
    super().__init__(name, handler)
    self.done = False
    self.frequency = frequency
    self.interval = interval
    self.loop = asyncio.get_event_loop()

  def transform_time(self, next_time):
    return get_rounded_timestamp(self.interval) - self.interval

  async def start(self):
    try:
      while not self.done:
        next_time = datetime.now().timestamp() + self.frequency
        next_time -= next_time % self.frequency
        await asyncio.sleep(next_time - datetime.now().timestamp())
        await self.emit_async('cron', self.transform_time(next_time))
    finally:
      self.done = True


class OffsetCronJob(CronJob):
  def __init__(self, name, handler,frequency, interval, offset):
    super().__init__(name, handler, frequency, interval)
    self.offset = offset

  def transform_time(self, next_time):
    return get_rounded_timestamp(self.interval) - self.interval - self.offset


class HistoryTimeEmitter(Processor):
  def __init__(self, name, handler, start, interval, end):
    super().__init__()
    self.done = False
    self.end_time = end
    self.handler = handler
    self.interval = interval
    self.name = name
    self.start_time = start

  async def start(self):
    try:
      current = self.start_time
      while current < self.end_time:
        await self.emit_async('cron', current)
        current += self.interval
    finally:
      self.done = True



#pylint: disable=E0401
"""
Test scheduler sources.
"""
import asyncio
from unittest import TestCase

from aiopype.sources import CronJob
from aiopype import SyncProtocol

class TestScheduler(TestCase):
  def test_scheduler(self):
    mock_handler = SyncProtocol()
    scheduler = CronJob('cron', mock_handler, frequency = 1, interval = 1)

    timestamps = []
    def handler(timestamp):
      timestamps.append(timestamp)
      if len(timestamps) == 3:
        scheduler.done = True

    scheduler.on('cron', handler)
    asyncio.get_event_loop().run_until_complete(scheduler.start())
    self.assertEqual(timestamps[0], timestamps[1] - 1)
    self.assertEqual(timestamps[1], timestamps[2] - 1)



"""
Main module

Contains the persister singleton.
"""
import asyncio
import logging
import sys

from raven import Client

from .manager import ManagerRegistry
from .protocol import AsyncProtocol, SyncProtocol


class Flow(object):
  """
  Persister entrypoint

  Responsible for orchestration, error handling and report.
  """
  def __init__(self, config):
    self.config = config
    self.logger = logging.getLogger('persister')
    if not hasattr(config, 'RAVEN_DSN'):
      config.RAVEN_DSN = None
    if config.RAVEN_DSN:
      self.raven = Client(config.RAVEN_DSN, release = config.VERSION)
    else:
      self.raven = None

    self.loop = asyncio.get_event_loop()
    self.loop.set_exception_handler(self.handle_exception)

  def handle_exception(self, _, data):
    if 'exception' in data:
      exception = data['exception']
      self.logger.warning(exception)

      exception_info = (type(exception), exception, exception.__traceback__)

      if self.raven:
        self.raven.captureException(exception_info)
      else:
        self.logger.info('Unable to capture exception on raven', extra = {'info': exception_info})

      self.logger.debug("Reported to sentry, will restart manager")
      asyncio.ensure_future(self.check_managers(), loop = self.loop)

  async def check_managers(self):
    self.logger.debug("Checking for stopped managers")

    done_managers = [manager for manager in self.managers if manager.done()]

    self.logger.debug('Stopped: ' + str(done_managers))

    for manager in done_managers:

      if manager.should_restart():
        manager.remove_all_listeners()
        factory = ManagerRegistry.REGISTRY[manager.name]
        new_manager = factory(self.handler)
        new_manager.start(self.loop)
        self.managers.append(new_manager)

      self.managers.remove(manager)

  async def finished(self):
    while self.managers:
      await asyncio.sleep(4)
      self.managers = [manager for manager in self.managers if manager.should_restart() or not manager.done()]
    self.logger.debug("All managers are done")

  def start(self):
    self.managers = []
    self.handler = SyncProtocol()
    flows = self.config.FLOWS

    for manager_factory in [ManagerRegistry.REGISTRY[x] for x in flows]:
      manager = manager_factory(self.handler)
      manager.start(self.loop)
      self.managers.append(manager)

    self.loop.run_until_complete(self.finished())


class AsyncFlow(Flow):
  async def check_managers(self):
    self.logger.debug("Checking for stopped managers")

    done_managers = [manager for manager in self.managers if manager.done()]

    self.logger.debug('Stopped: ' + str(done_managers))

    for manager in done_managers:

      if manager.should_restart():
        manager.remove_all_listeners()
        factory = ManagerRegistry.REGISTRY[manager.name]
        new_manager = factory(self.handler)
        new_manager.start(self.loop)
        self.managers.append(new_manager)

      self.managers.remove(manager)

    if not self.handler.running:
      self.logger.error("Listener stopped, restarting.")
      self.handler_task = asyncio.ensure_future(self.check_listener())

  async def check_listener(self):
    await self.handler.listen()

  async def finished(self):
    await super().finished()
    await self.handler.emit_async('stop')
    await self.handler_task
    self.logger.debug("Messaging handler done")

  def start(self):
    self.managers = []
    self.handler = AsyncProtocol()
    flows = self.config.FLOWS

    for manager_factory in [ManagerRegistry.REGISTRY[x] for x in flows]:
      manager = manager_factory(self.handler)
      manager.start(self.loop)
      self.managers.append(manager)

    self.handler_task = asyncio.ensure_future(self.check_listener(), loop = self.loop)
    self.loop.run_until_complete(self.finished())



"""
Aiopype - Flow based programming with asyncio.
"""

from .flow import Flow, AsyncFlow
from .manager import Manager, DescriptiveManager
from .processor import Processor
from .protocol import AsyncProtocol, SyncProtocol
